/*
 * For a copy of the 20080320 version, load http://www.blackcore.net/source/puttymonitor-20080320.c
 *
 * puttymonitor is a network diagnostic application to determine
 * lag over a terminal. This application sends character 5 (^E)
 * which triggers an auto-response from PuTTY. This allows the
 * application to determine the latency through an active connection.
 * ncurses is required to compile this application.
 * 
 * Terminal->Answerback to ^E should have PuTTY for this application to work.
 *
 * gcc -o puttymonitor puttymonitor.c -Wall -lcurses
 *
 * Press F1 to print help (this pauses application)
 * Press F2 to display statistics
 * Press F3 to quit application
 * Press F4 to pause application
 * 20080426
 * - Changed verious function and variable names.
 * - Added getopt_long() support, application will now take command line arguments
 * - select() support added
 * - Minimum and maximum statistics have been added.
 * - Average now works properly (before it would display whole numbers trailing always with .00)
 * - Packet loss emulation added. If packets exceed config.interval*1000 it is assumed a lost packet
 *    It is done this way because the packets will still be received regardless because TCP is a
 *    connection oriented protocol. If anyone has a better way, find a way to contact me.
 * - Changed command keys for statistics and quitting; F2 shows statistics, F3 quits
 * - Added additional command keys, F1 shows in application help (curses) and F4 pauses application
 *    This fixes the [returnstr] bug; you could not define a returnstr with S and Q before
 * - [returnstr] variable is now called config.expect. Environment variable RETURNSTR is deprecated,
 *    use PM_EXPECT instead.
 * - Added various comments to make source easier to understand
 * - Removed SIGALRM signal since we are using select() now
 * - Added SIGQUIT signal allowing CTRL+\ input to display statistics
 * - Exit code 127 indicates that finding a suitable print function has failed
 * 20080320
 * - Changed RETURNSTR to [char type returnstr]. You can now define environment variable RETURNSTR 
 *    for an alternate string other than PuTTY
  
    $Id: puttymonitor.c,v 0 2007/03/27 10:08:00 -8 Exp $ Copyright 2007-2008, blackcore networking
    Copyright 2007-2008, blackcore networking.
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curses.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#define __GNU_SOURCE
#include <getopt.h>
/* Definitions  */

#define MAX_STRING 128
// Verbosity
#define DEBUG_LOW 1
#define DEBUG_MED 5
#define DEBUG_HIGH 9
// Flags
#define FLAG_WAITING 0x00000001 // This isn't used right now
#define FLAG_INPACKET 0x00000002 // We are currently in a packet
#define FLAG_HAVEPUTTY 0x00000004 // Application has full [returnstr]
#define FLAG_PAUSED 0x00000008 // Pause application
/* End definitions */

// Function declarations
void signal_handler(int);

// print_stats() shows statistics
void print_stats();

// do_exit(exitcode) exits application
void do_exit(int);

// getnow() will return tv_sec+tv_usec so that we can get a millisecond readout
unsigned long long getnow();

// send_ping() sends ^E to the terminal among other things
void send_ping();

// print_help() should be pretty straight forward. 0 for console, 1 for ncurses
void print_help(int);

// initapp() is to make it so that main() isn't so cluttered
int initapp();

// get_options() is the getopt_long section
int get_options(int argc, char *argv[]);

// print_verbose() calls printw() if we are verbose enough
int print_verbose(unsigned long, char *, ...);

// toggle_pause()
void toggle_pause(int);

// End function declarations

// Regular declarations
unsigned long flags = 0;
unsigned long long time_sent;
unsigned long total = 0, sent = 0, recv = 0;
unsigned long stats_max = 0, stats_min = 0;

struct Config {
	char expect[MAX_STRING]; // This is returnstr
	int interval; // How often we are expecting to send packets and wait on them
	int running;
	unsigned long verbosity;
};
struct HelpIndex { 
	char *key;
	char *value;
};

static struct Config config;

extern char *optarg;
extern int optind, opterr, optopt;

// End regular declarations
int main(int argc, char *argv[]) {
	int c, retval;
	char sa[2], check[MAX_STRING];
	struct timeval tv;
	fd_set rfds;

	memset(&config, 0, sizeof(config));

	if (get_options(argc, argv)) {
		return 2;
	}

	switch(initapp()) {
		case 1:
			printf("A fatal error has occured. puttymonitor could not load ncurses.\n");
			printf("Make sure that your environment variable TERM is set properly.\n");
			printf("export TERM=vt100 # bash syntax -- will set terminal to vt100 emulation\n");
			return 1;
		break;
	}

// Add signal event for SIGINT (signal interrupt), also known as ^C (exits application)
	signal(SIGINT, &signal_handler);
// Add signal event for SIGQUIT. This is called when CTRL+\ is pressed. (displays statistics);
	signal(SIGQUIT, &signal_handler);

	memset(check, 0, sizeof(check));
	send_ping(); // Send our initial ping so we don't have to wait
	if (flags & FLAG_PAUSED)
		toggle_pause(0);
	while (config.running) {
		FD_ZERO(&rfds);
		FD_SET(0, &rfds);
		tv.tv_sec = config.interval;
		tv.tv_usec = 0;
		retval = select(1, &rfds, NULL, NULL, &tv);
		if (retval <= 0) {
			if (flags & FLAG_PAUSED)
				continue;
			// If it has been one second and we are still waiting then the packet has been lost.
			// Send the packet again
			if (!(flags & FLAG_HAVEPUTTY))
				printw("no reply!\n");

			memset(check, 0, sizeof(check));
			send_ping();
			continue;
		}
		c = getch();
		refresh();
		switch(c) {
			case KEY_F(4):
				toggle_pause(1);
				break;
			// F3 quits application
			case KEY_F(3):
				do_exit(0);
				break;
			// F2 shows statistics
			case KEY_F(2):
				print_stats();
				break;
			case KEY_F(1):
				print_help(1);
				break;
			default:
				// Only append data if we are NOT paused
				if (!(flags & FLAG_PAUSED)) {
					int l;
					l = strlen(check); // Set current position into l
					if (config.expect[l] != c)
						print_verbose(DEBUG_LOW, "config.expect[%d](%c) != %c\n", l, config.expect[l], c);
					else {
						sa[0] = c;
						sa[1] = 0;
						flags |= FLAG_INPACKET;
						strncat(check, sa, sizeof(check) - 1);
					}
				}
				break;
		}
		if (!strcmp(check, config.expect)) {
			time_sent = getnow() - time_sent;
// If it takes longer than 1000ms to reply, we consider it a lost packet
// Considering it has been more than one second the application will have moved
// On anyhow
			if (time_sent > (config.interval * 1000)) {
//				printw("(LOST) at %llums", time_sent);
				flags &= ~(FLAG_HAVEPUTTY|FLAG_WAITING);
				continue;
			}
			printw("%s: %llums\n", check, time_sent);
			if (!stats_min)
				stats_min = time_sent;
			if (time_sent > stats_max) stats_max = time_sent;
			if (time_sent < stats_min) stats_min = time_sent;
			total+=time_sent;
			recv++;
			flags &= ~FLAG_WAITING;
			flags |= FLAG_HAVEPUTTY;
			memset(check, 0, sizeof(check));
		}
		refresh();
	}
	endwin();
	return 0;
}
void print_stats() {
	unsigned long lost;
	float avg;
	float loss;
	if (sent == 0) {
		printw("No statistics are available at this time.\n");
		return;
	}
	lost = sent - recv;
	avg = (float)((float)total / (float)sent);
	loss = (float)((float)lost / (float)sent) * 100.0;
	printw("sent=%lu, recv=%lu, lost=%lu loss=%.2f%% ", sent, recv, lost, loss);
	printw("min=%lu avg=%.2fms max=%lu\n", stats_min, avg, stats_max);
	refresh();
}
void do_exit(int exitcode) {
	print_stats();
	refresh();
	endwin();
	_exit(exitcode);
}
void signal_handler(int sig) {
	switch(sig) {
	case SIGINT: // signal interrupt, ^C
		do_exit(2);
		break;
	case SIGQUIT: // signal quit, show statistics
		print_stats();
		break;
	default:
		printw("Invalid signal caught: %d\n", sig);
		break;
	}
}
void send_ping() {
	// If we are paused we won't send data
	if (flags & FLAG_PAUSED)
		return;
	/*
	 * ASCII 5 causes PuTTY to respond with 'PuTTY'
	 * We have to use printf() because ncurses prints ^E
	*/
	time_sent = getnow();
	flags |= FLAG_WAITING;
	flags &= ~FLAG_HAVEPUTTY;
	printf("\x05");
	fflush(stdout);
	printw("Sending request... ");
	sent++;
	refresh();
}
unsigned long long getnow() {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return ((tv.tv_sec * 1000) + (tv.tv_usec / 1000));
	
}
/* 
 * Welcome to my sloppy help function!
 * If this doesn't work on your platform for some reason, let me know and I might fix it
*/
void print_help(int type) {
	int i;
	typedef int (*print_t)(const char *, ...);
	static print_t printfunc = NULL;
	struct HelpIndex helpindex_console[] = {
		{"--help",		"Displays console help"},
		{"--ncurseshelp"	"Displays ncurses help"},
		{"--expect",		"Defines expect variable (will override PM_EXPECT)"},
		{"--pause",		"Starts application paused"},
		{"--verbosity",		"Sets verbosity level (0 is off, 9 is highest)"},
		{"--interval",		"Changes interval in seconds (default is 1)"},
		{NULL, NULL}
	};
	struct HelpIndex helpindex_ncurses[] = {
		{"^C", "Displays statistics and exits application"},
		{"F1", "Show help"},
		{"F2", "Show statistics"},
		{"F3", "Displays statistics and exits application"},
		{"F4", "Pause/Unpause application"},
		{NULL, NULL}
	};
	struct HelpIndex *h = NULL;

	if (config.running)
		printfunc = (print_t)printw;
	else
		printfunc = (print_t)printf;
	if (printfunc == NULL) {
		_exit(127);
	}
	switch(type) {
	case 1: // ncurses help, using printw()
		h = helpindex_ncurses;
		break;
	case 0:
		h = helpindex_console;
		break;
	}
	if (h)
		for(i=0;h[i].key != NULL;i++)
			printfunc("%-20s %s\n", h[i].key, h[i].value);
	else
		printfunc("Error displaying help!\n");

}
int initapp() {
	config.interval = 1;

	// RETURNSTR is deprecated, use PM_EXPECT
	if (!strlen(config.expect)) {
		if (getenv("PM_EXPECT"))
			strncpy(config.expect, getenv("PM_EXPECT"), sizeof(config.expect));
		else if (getenv("RETURNSTR"))
			strncpy(config.expect, getenv("RETURNSTR"), sizeof(config.expect));
		else
			strncpy(config.expect, "PuTTY", sizeof(config.expect));
	}
	// Init ncurses
	if (initscr() == NULL)
		return 1;
	keypad(stdscr, TRUE); // This allows F1, F2 and F3 usage
	noecho();
	scrollok(stdscr, TRUE);
	/*
	 * At this point refresh() will clear the screen. This is added because FLAG_PAUSED will
	 * wait on a key press before refreshing the screen.
	*/
	refresh();
	config.running = 1;
	return 0;
}
int get_options(int argc, char *argv[]) {
	int c, a;
//	int digit_optind;
	for(;;) {
//		int this_option_optind = optind ? optind : 1;
		int option_index = 0;
		static struct option long_options[] = {
			{"help",	0, 0, 0},
			{"ncurseshelp",	0, 0, 0},
			{"expect",	required_argument, 0, 0},
			{"verbosity",	required_argument, 0, 0},
			{"pause",	0, 0, 0},
			{"interval",	required_argument, 0, 0},
			{0, 0, 0, 0}
		};
		c = getopt_long(argc, argv, "e:", long_options, &option_index);
		if (c == -1)
			break;
		switch(c) {
		case 0:
			if (!strcmp(long_options[option_index].name, "help")) {
				print_help(0);
				return 1;
			} else if (!strcmp(long_options[option_index].name, "ncurseshelp")) {
				print_help(1);
				return 1;
			} else if (!strcmp(long_options[option_index].name, "expect")) {
				strncpy(config.expect, optarg, sizeof(config.expect));
			} else if (!strcmp(long_options[option_index].name, "verbosity")) {
				a = atoi(optarg);
				if (a >= 0 && a <= 9)
					config.verbosity = a;
				else
					config.verbosity = 0;
			} else if (!strcmp(long_options[option_index].name, "pause")) {
				flags |= FLAG_PAUSED;
			} else if (!strcmp(long_options[option_index].name, "interval")) {
				a = atoi(optarg);
				if (a > 120) {
					printf("You may have made a mistake! You set your interval higher than two minutes. (%.2f minutes)\n", (float)((float)a / (float)60));
					printf("Sleeping for 2 seconds before starting application...\n");
					sleep(2);
				} else if (a < 0) {
					printf("interval is a negative number\n");
					return 2;
				} else
					config.interval = a;
			}
			break;
		case 'e':
			
			break;
		default:
			printf ("?? getopt returned character code 0%o ??\n", c);
			return 3;
		}
	}
	if (optind < argc) {
		while(optind < argc) 
			printf("%s ", argv[optind++]);
		printf("\n");
	}
	return 0;	
}
int print_verbose(unsigned long verbosity, char *fmt, ...) {
	int n;
	char buf[MAX_STRING];
	va_list vargs;

	// Check if our verbose level is high enough to print
	if (config.verbosity < verbosity) return -1;

	va_start(vargs, fmt);
	n=vsnprintf(buf, sizeof(buf), fmt, vargs);
	printw(buf);
	refresh();
	va_end(vargs);

	return n;
}
void toggle_pause(int type) {
	// If type 0 is passed, just print the message.
	if (type == 1) {
		flags ^= FLAG_PAUSED;
		// FLAG_HAVEPUTTY is set so that the "no reply" message does not display
		flags |= FLAG_HAVEPUTTY;
	}
	printw("Application is %s paused.\n", (flags & FLAG_PAUSED)?"now":"no longer");
	refresh();
}
